@@ -16,10 +16,11 @@
 #include <sched.h>
 #include <sys/socket.h>
 
+#include "zdtmtst.h"
 #include "ns.h"
 
 extern int pivot_root(const char *new_root, const char *put_old);
-static int prepare_mntns()
+static int prepare_mntns(void)
 {
 	int dfd, ret;
 	char *root;
@@ -31,115 +32,110 @@ static int prepare_mntns()
 		return -1;
 	}
 
-		/*
-		 * In a new userns all mounts are locked to protect what is
-		 * under them. So we need to create another mount for the
-		 * new root.
-		 */
-		if (mount(root, root, NULL, MS_SLAVE , NULL)) {
-			fprintf(stderr, "Can't bind-mount root: %m\n");
-			return -1;
-		}
+	/*
+	 * In a new userns all mounts are locked to protect what is
+	 * under them. So we need to create another mount for the
+	 * new root.
+	 */
+	if (mount(root, root, NULL, MS_SLAVE , NULL)) {
+		fprintf(stderr, "Can't bind-mount root: %m\n");
+		return -1;
+	}
 
-		if (mount(root, root, NULL, MS_BIND | MS_REC, NULL)) {
-			fprintf(stderr, "Can't bind-mount root: %m\n");
-			return -1;
-		}
+	if (mount(root, root, NULL, MS_BIND | MS_REC, NULL)) {
+		fprintf(stderr, "Can't bind-mount root: %m\n");
+		return -1;
+	}
 
-		/* Move current working directory to the new root */
-		ret = readlink("/proc/self/cwd", path, sizeof(path) - 1);
-		if (ret < 0)
-			return -1;
-		path[ret] = 0;
+	/* Move current working directory to the new root */
+	ret = readlink("/proc/self/cwd", path, sizeof(path) - 1);
+	if (ret < 0)
+		return -1;
+	path[ret] = 0;
 
-		dfd = open(path, O_RDONLY | O_DIRECTORY);
-		if (dfd == -1) {
-			fprintf(stderr, "open(.) failed: %m\n");
-			return -1;
-		}
+	dfd = open(path, O_RDONLY | O_DIRECTORY);
+	if (dfd == -1) {
+		fprintf(stderr, "open(.) failed: %m\n");
+		return -1;
+	}
 
-		if (chdir(root)) {
-			fprintf(stderr, "chdir(%s) failed: %m\n", root);
-			return -1;
-		}
-		if (mkdir("old", 0777) && errno != EEXIST) {
-			fprintf(stderr, "mkdir(old) failed: %m\n");
-			return -1;
-		}
+	if (chdir(root)) {
+		fprintf(stderr, "chdir(%s) failed: %m\n", root);
+		return -1;
+	}
+	if (mkdir("old", 0777) && errno != EEXIST) {
+		fprintf(stderr, "mkdir(old) failed: %m\n");
+		return -1;
+	}
 
-		if (pivot_root(".", "./old")) {
-			fprintf(stderr, "pivot_root(., ./old) failed: %m\n");
-			return -1;
-		}
+	if (pivot_root(".", "./old")) {
+		fprintf(stderr, "pivot_root(., ./old) failed: %m\n");
+		return -1;
+	}
 
-		if (mount("./old", "./old", NULL, MS_PRIVATE | MS_REC , NULL)) {
-			fprintf(stderr, "Can't bind-mount root: %m\n");
-			return -1;
-		}
+	if (mount("./old", "./old", NULL, MS_PRIVATE | MS_REC , NULL)) {
+		fprintf(stderr, "Can't bind-mount root: %m\n");
+		return -1;
+	}
 
-		if (mkdir("proc", 0777) && errno != EEXIST) {
-			fprintf(stderr, "mkdir(proc) failed: %m\n");
-			return -1;
-		}
+	/*
+	 * proc and sysfs can be mounted in an unprivileged namespace,
+	 * if they are already mounted when the user namespace is created.
+	 * So ./old must be umounted after mounting /proc and /sys.
+	 */
+	if (mount("proc", "/proc", "proc", MS_MGC_VAL | MS_NOSUID | MS_NOEXEC | MS_NODEV, NULL)) {
+		fprintf(stderr, "mount(/proc) failed: %m\n");
+		return -1;
+	}
 
-		/*
-		 * proc and sysfs can be mounted in an unprivileged namespace,
-		 * if they are already mounted when the user namespace is created.
-		 * So ./old must be umounted after mounting /proc and /sys.
-		 */
-		if (mount("proc", "/proc", "proc", MS_MGC_VAL | MS_NOSUID | MS_NOEXEC | MS_NODEV, NULL)) {
-			fprintf(stderr, "mount(/proc) failed: %m\n");
-			return -1;
-		}
+	if (umount2("./old", MNT_DETACH)) {
+		fprintf(stderr, "umount(./old) failed: %m\n");
+		return -1;
+	}
 
-		if (umount2("./old", MNT_DETACH)) {
-			fprintf(stderr, "umount(./old) failed: %m\n");
+	if (mount("pts", "/dev/pts", "devpts", MS_MGC_VAL, "mode=666,ptmxmode=666,newinstance")) {
+		fprintf(stderr, "mount(/dev/pts) failed: %m\n");
+		return -1;
+	}
+	/*
+	 * If CONFIG_DEVPTS_MULTIPLE_INSTANCES=n, then /dev/pts/ptmx
+	 * does not exist. Fall back to creating the device with
+	 * mknod() in that case.
+	 */
+	if (access("/dev/pts/ptmx", F_OK) == 0) {
+		if (symlink("pts/ptmx", "/dev/ptmx") && errno != EEXIST) {
+			fprintf(stderr, "symlink(/dev/ptmx) failed: %m\n");
 			return -1;
 		}
+	} else {
+		if (mknod("/dev/ptmx", 0666 | S_IFCHR, makedev(5, 2)) == 0) {
+			chmod("/dev/ptmx", 0666);
+		} else if (errno != EEXIST) {
+			fprintf(stderr, "mknod(/dev/ptmx) failed: %m\n");
+			return -1;
+		}
+	}
 
-		if (mkdir("/dev", 0755) && errno != EEXIST) {
-			fprintf(stderr, "mkdir(/dev) failed: %m\n");
-			return -1;
-		}
-		if (mkdir("/dev/pts", 0755) && errno != EEXIST) {
-			fprintf(stderr, "mkdir(/dev/pts) failed: %m\n");
-			return -1;
-		}
-		if (mount("pts", "/dev/pts", "devpts", MS_MGC_VAL, "mode=666,ptmxmode=666,newinstance")) {
-			fprintf(stderr, "mount(/dev/pts) failed: %m\n");
-			return -1;
-		}
-		/*
-		 * If CONFIG_DEVPTS_MULTIPLE_INSTANCES=n, then /dev/pts/ptmx
-		 * does not exist. Fall back to creating the device with
-		 * mknod() in that case.
-		 */
-		if (access("/dev/pts/ptmx", F_OK) == 0) {
-			if (symlink("pts/ptmx", "/dev/ptmx") && errno != EEXIST) {
-				fprintf(stderr, "symlink(/dev/ptmx) failed: %m\n");
-				return -1;
-			}
-		} else {
-			if (mknod("/dev/ptmx", 0666 | S_IFCHR, makedev(5, 2)) == 0) {
-				chmod("/dev/ptmx", 0666);
-			} else if (errno != EEXIST) {
-				fprintf(stderr, "mknod(/dev/ptmx) failed: %m\n");
-				return -1;
-			}
-		}
-		if (access("/dev/tty", F_OK)) {
-			if (mknod("/dev/tty", 0666 | S_IFCHR, makedev(5, 0)) == 0) {
-				chmod("/dev/tty", 0666);
-			} else if (errno != EEXIST) {
-				fprintf(stderr, "mknod(/dev/tty) failed: %m\n");
-				return -1;
-			}
-		}
-		if (fchdir(dfd)) {
-			fprintf(stderr, "fchdir() failed: %m\n");
-			return -1;
-		}
-		close(dfd);
+	if (fchdir(dfd)) {
+		fprintf(stderr, "fchdir() failed: %m\n");
+		return -1;
+	}
+	close(dfd);
+
+	return 0;
+}
+
+static int prepare_namespaces(void)
+{
+	if (setuid(0) || setgid(0) || setgroups(0, NULL)) {
+		fprintf(stderr, "set*id failed: %m\n");
+		return -1;
+	}
+
+	system("ip link set up dev lo");
+
+	if (prepare_mntns())
+		return -1;
 
 	return 0;
 }
@@ -148,7 +144,7 @@ static int prepare_mntns()
 
 /* All arguments should be above stack, because it grows down */
 struct ns_exec_args {
-	char stack[NS_STACK_SIZE];
+	char stack[NS_STACK_SIZE] __stack_aligned__;
 	char stack_ptr[0];
 	int argc;
 	char **argv;
@@ -195,7 +191,7 @@ write_out:
 }
 
 #define STATUS_FD 255
-int ns_exec(void *_arg)
+static int ns_exec(void *_arg)
 {
 	struct ns_exec_args *args = (struct ns_exec_args *) _arg;
 	char buf[4096];
@@ -213,14 +209,8 @@ int ns_exec(void *_arg)
 	close(args->status_pipe[1]);
 	read(STATUS_FD, buf, sizeof(buf));
 	shutdown(STATUS_FD, SHUT_RD);
-	if (setuid(0) || setgid(0) || setgroups(0, NULL)) {
-		fprintf(stderr, "set*id failed: %m\n");
-		return -1;
-	}
 
-	system("ip link set up dev lo");
-
-	if (prepare_mntns())
+	if (prepare_namespaces())
 		return -1;
 
 	setenv("ZDTM_NEWNS", "2", 1);
@@ -229,6 +219,19 @@ int ns_exec(void *_arg)
 	return -1;
 }
 
+static void show_ps(void)
+{
+	int pid;
+
+	pid = fork();
+	if (pid == 0) {
+		execl("/bin/ps", "ps", "axf", "-o", "pid,sid,comm", NULL);
+		fprintf(stderr, "Unable to execute ps: %m\n");
+		exit(1);
+	} else if (pid > 0)
+		waitpid(pid, NULL, 0);
+}
+
 int ns_init(int argc, char **argv)
 {
 	struct sigaction sa = {
@@ -260,10 +263,9 @@ int ns_init(int argc, char **argv)
 		fprintf(stderr, "fork() failed: %m\n");
 		exit(1);
 	} else if (pid == 0) {
-		setenv("ZDTM_NEWNS", "3", 1);
-		ret = execvp(argv[0], argv);
-		fprintf(stderr, "exec(%s) failed: %m\n", argv[0]);
-		return ret;
+		close(status_pipe);
+		unsetenv("ZDTM_NEWNS");
+		return 0; /* Continue normal test startup */
 	}
 
 	ret = -1;
@@ -272,13 +274,7 @@ int ns_init(int argc, char **argv)
 	else if (ret)
 		fprintf(stderr, "The test returned non-zero code %d\n", ret);
 
-	pid = fork();
-	if (pid == 0) {
-		execl("/bin/ps", "ps", "axf", "-o", "pid,sid,comm", NULL);
-		fprintf(stderr, "Unable to execute ps: %m\n");
-		exit(1);
-	} else if (pid > 0)
-		waitpid(pid, NULL, 0);
+	show_ps();
 
 	if (sigaction(SIGCHLD, &sa, NULL)) {
 		fprintf(stderr, "Can't set SIGCHLD handler: %m\n");
@@ -303,18 +299,12 @@ int ns_init(int argc, char **argv)
 	write(status_pipe, &ret, sizeof(ret));
 	close(status_pipe);
 	if (ret)
-		return ret;
+		exit(ret);
 
 	/* suspend/resume */
 	test_waitsig();
 
-	pid = fork();
-	if (pid == 0) {
-		execl("/bin/ps", "ps", "axf", "-o", "pid,sid,comm", NULL);
-		fprintf(stderr, "Unable to execute ps: %m\n");
-		exit(1);
-	} else if (pid > 0)
-		waitpid(pid, NULL, 0);
+	show_ps();
 
 	fd = open(pidfile, O_RDONLY);
 	if (fd == -1) {
@@ -340,46 +330,6 @@ int ns_init(int argc, char **argv)
 	exit(1);
 }
 
-static int construct_root()
-{
-	char *root;
-	int dfd;
-
-	root = getenv("ZDTM_ROOT");
-	if (!root) {
-		fprintf(stderr, "ZDTM_ROOT isn't set\n");
-		return -1;
-	}
-
-	dfd = open(".", O_RDONLY);
-	if (dfd == -1) {
-		fprintf(stderr, "open(.) failed: %m\n");
-		return -1;
-	}
-	if (chdir(root)) {
-		fprintf(stderr, "chdir(%s): %m\n", root);
-		return -1;
-	}
-
-	mkdir("dev", 0777);
-	chmod("dev", 0777);
-	mknod("dev/null", 0777 | S_IFCHR, makedev(1, 3));
-	chmod("dev/null", 0777);
-	mkdir("dev/net", 0777);
-	mknod("dev/net/tun", 0777 | S_IFCHR, makedev(10, 200));
-	chmod("dev/net/tun", 0777);
-	mknod("dev/rtc", 0777 | S_IFCHR, makedev(254, 0));
-	chmod("dev/rtc", 0777);
-
-	if (fchdir(dfd)) {
-		fprintf(stderr, "fchdir() failed: %m\n");
-		return -1;
-	}
-	close(dfd);
-
-	return 0;
-}
-
 #define UID_MAP "0 100000 100000\n100000 200000 50000"
 #define GID_MAP "0 400000 50000\n50000 500000 100000"
 void ns_create(int argc, char **argv)
@@ -400,20 +350,12 @@ void ns_create(int argc, char **argv)
 		exit(1);
 	}
 
+	flags = CLONE_NEWPID | CLONE_NEWNS | CLONE_NEWUTS |
+		CLONE_NEWNET | CLONE_NEWIPC | SIGCHLD;
+
 	val = getenv("ZDTM_USERNS");
 	if (val)
-		/*
-		 * CLONE_NEWIPC and CLONE_NEWUTS are excluded, because
-		 * their sysctl-s are protected by CAP_SYS_ADMIN
-		 */
-		flags = CLONE_NEWPID | CLONE_NEWNS  |
-			CLONE_NEWNET | CLONE_NEWUSER | SIGCHLD;
-	else
-		flags = CLONE_NEWPID | CLONE_NEWNS | CLONE_NEWUTS |
-			CLONE_NEWNET | CLONE_NEWIPC | SIGCHLD;
-
-	if (construct_root())
-		exit(1);
+		flags |= CLONE_NEWUSER;
 
 	pid = clone(ns_exec, args.stack_ptr, flags, &args);
 	if (pid < 0) {
@@ -423,7 +365,7 @@ void ns_create(int argc, char **argv)
 
 	close(args.status_pipe[1]);
 
-	if (val) {
+	if (flags & CLONE_NEWUSER) {
 		snprintf(pname, sizeof(pname), "/proc/%d/uid_map", pid);
 		fd = open(pname, O_WRONLY);
 		if (fd < 0) {
